clc
clearvars
load MeshFiles
mesh = LShape;          % [-1,1]^2\[0,1]^2
%mesh = SquareT;        % [0,1]^2 with regular mesh
%mesh = SquareTFine;    % [0,1]^2 but with less regular mesh

%%%%% uncomment the function that you want as exact solution
%%% custom symbolic function
% syms x y
% u = cos(2*x).*sin(3*y);   % <-- change this!
% q = matlabFunction(gradient(u),'vars',{x,y});
% f = matlabFunction(-laplacian(u),'vars',{x,y});
% u = matlabFunction(u,'vars',{x,y});
% clear x y

%%% harmonic r^(2/3)*sin((2*theta - pi)/3)) for the LShape
u = @(x,y) uD(x,y);
q = @(x,y) [2/3*(x.^2+y.^2).^(-2/3).*x.*sin((2*atan3(y,x)-pi)/3) + (x.^2+y.^2).^(1/3).*cos((2*atan3(y,x)-pi)/3).*(-2*y./(3*(x.^2+y.^2)));
    2/3*(x.^2+y.^2).^(-2/3).*y.*sin((2*atan3(y,x)-pi)/3) + (x.^2+y.^2).^(1/3).*cos((2*atan3(y,x)-pi)/3).*(2*x./(3*(x.^2+y.^2)))];
f = @(x,y) (x-x) + (y-y);
for i = 1:length(mesh)
    coord = mesh(i).coordinates;
    elem = mesh(i).elements;
    [U,errors] = FEMDirichletSolver(f,u,q,coord,elem);
    [h(i),L2(i),H1(i)] = errors{:};
    
    figure(1)
    tiledlayout(2,1)
    ax1 = nexttile;
    trisurf(elem,coord(:,1),coord(:,2),U,'linewidth',0.5)
    title(sprintf('Numerical solution with N = %i elements',size(elem,1)));
    ax2 = nexttile;
    trisurf(elem,coord(:,1),coord(:,2),u(coord(:,1),coord(:,2))); shading interp
    linkprop([ax1, ax2],{'CameraUpVector', 'CameraPosition', 'CameraTarget', 'XLim', 'YLim', 'ZLim'});
    colormap jet
    colorbar
    view(3)
    drawnow
    pause
end
%%
figure(2)
loglog(h,L2,'xb'); hold on
loglog(h,H1,'or');
p = polyfit(log(h),log(L2),1);
loglog(h,exp(p(2))*h.^p(1),'b--');
p = polyfit(log(h),log(H1),1);
loglog(h,exp(p(2))*h.^p(1),'r--'); hold off
grid on
legend('L2 error','H1 error');
set(gca,'xdir','reverse')
axis equal
title('Error convergence');

clc
printdata = [round(1./h.^2); 
    L2;0 log(L2(2:end)./L2(1:end-1))./log(h(2:end)./h(1:end-1));
    H1;0 log(H1(2:end)./H1(1:end-1))./log(h(2:end)./h(1:end-1))];
fprintf('-----------|      Estimated Convergence     |----------\n');
fprintf('| Nelts |   L^2 error  | e.o.c.|   H^1 error  | e.o.c.|\n');
fprintf('|%7i|%12.2e  |%6.2f |%12.2e  |%6.2f |\n',printdata)
fprintf('-------------------------------------------------------\n');

%% ADDITIONAL FUNCTIONS
function y = uD(x,y)
r = sqrt(x.^2+y.^2);
th = atan3(y,x);
y = r.^(2/3).*sin((2*th-pi)/3);
end
function th = atan3(y,x)
th = atan2(y,x);
th(th <= 0) = th(th <= 0) + 2*pi;
end

function [U,errors] = FEMDirichletSolver(f,uD,qN,coord,elem)
%%%%% SOLVING THE GALERKIN SYSTEM
DOF = size(coord,1);
% obtaining matrices
[MG, KG] = FEM2DLagrangeP1(coord,elem);
% boundary conditions
U = zeros(DOF,1);
DirNodes = find(coord(:,3) == 1);
FreeNodes = setdiff(1:DOF,DirNodes);
U(DirNodes) = uD(coord(DirNodes,1),coord(DirNodes,2));
%%% SOLVING LINEAR SYSTEM
F = MG*f(coord(:,1),coord(:,2));
b = (F - KG*U);
U(FreeNodes) = KG(FreeNodes,FreeNodes)\b(FreeNodes);
%%% ERRORS
L2 = errorl2(U,uD,coord,elem);
errors = {1/sqrt(size(elem,1)),...
          sqrt(L2),...
          sqrt(L2 + errorh1(U,qN,coord,elem))};
end

function L2 = errorl2(U,uD,coord,elem)
% L2 SQUARED NORM
%     U: solution vector
%    uD: exact solution (anonymous function)
% coord: coordinates (Nelem x 4) 
%  elem: connectivity matrix (Nelem x 3)

coordX = coord(:,1); coordY = coord(:,2);
X = coordX(elem); Y = coordY(elem);
detB = X(:,1).*Y(:,2) - X(:,2).*Y(:,1) - X(:,1).*Y(:,3) + ...
    X(:,3).*Y(:,1) + X(:,2).*Y(:,3) - X(:,3).*Y(:,2);

baryX = sum(X,2)/3;
baryY = sum(Y,2)/3;

uvalues = uD(baryX,baryY);
uh = sum(U(elem),2)/3;

E = (uh - uvalues).^2;
L2 = sum(E.*detB/2);
end

function H1 = errorh1(U,qN,coord,elem)
% L2 SQUARED NORM
%     U: solution vector
%    qN: exact solution (anonymous function)
% coord: coordinates (Nelem x 4) 
%  elem: connectivity matrix (Nelem x 3)

coordX = coord(:,1); coordY = coord(:,2);
X = coordX(elem); Y = coordY(elem);
detB = X(:,1).*Y(:,2) - X(:,2).*Y(:,1) - X(:,1).*Y(:,3) + ...
    X(:,3).*Y(:,1) + X(:,2).*Y(:,3) - X(:,3).*Y(:,2);

baryX = sum(X,2)/3;
baryY = sum(Y,2)/3;

d_uvalues = qN(baryX',baryY')';
A = U(elem);
A = (A(:,2:3) - A(:,1));

c11 = (Y(:,3) - Y(:,1))./detB;
c12 = (Y(:,1) - Y(:,2))./detB;
c21 = (X(:,1) - X(:,3))./detB;
c22 = (X(:,2) - X(:,1))./detB;

d_uh = [c11.*A(:,1) + c12.*A(:,2), c21.*A(:,1) + c22.*A(:,2)];

E = d_uh - d_uvalues;
H1 = sum(diag(E*E').*detB/2);
end
